import datetime
from enum import Enum
from typing import ClassVar, Generic, Literal, TypeGuard, TypeVar

from pydantic import BaseModel, Field


class UseToolsMode(str, Enum):
    read_only = "read_only"
    read_write = "read_write"
    disabled = "disabled"


class UseToolsConfig(BaseModel):
    mode: UseToolsMode = UseToolsMode.read_only

    @property
    def read_only(self) -> bool:
        return self.mode == UseToolsMode.read_only

    @property
    def read_write(self) -> bool:
        return self.mode == UseToolsMode.read_write

    def disable(self) -> None:
        self.mode = UseToolsMode.disabled


class CreateChatResponse(BaseModel):
    chat_uuid: str


class StartChatRequest(BaseModel):
    chat_uuid: str
    prompt: str
    use_tools_config: UseToolsConfig


class StartChatResponse(BaseModel):
    chat_uuid: str


class ExecutionEndResponse(BaseModel):
    execution_ended: bool


class GitInfo(BaseModel):
    branch: str
    remote: str | None


class PythonEnvInfo(BaseModel):
    interpreter_path: str | None
    interpreter_version: str | None
    name: str | None = "exponent"
    provider: Literal["venv", "pyenv", "pipenv", "conda"] | None = "pyenv"


class SystemInfo(BaseModel):
    name: str
    cwd: str
    os: str
    shell: str
    git: GitInfo | None
    python_env: PythonEnvInfo | None


class HeartbeatInfo(BaseModel):
    system_info: SystemInfo | None
    timestamp: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.now(datetime.UTC)
    )


Direction = Literal[
    "request",
    "response",
]

Namespace = Literal[
    "code_execution",
    "list_files",
    "get_file_contents",
    "get_matching_files",
    "system_context",
    "get_all_tracked_files",
]

SupportedLanguage = Literal[
    "python",
    "shell",
    "diff",
]


class RemoteExecutionMessageData(BaseModel):
    namespace: Namespace
    direction: Direction
    message_data: str

    def message_type(self) -> str:
        return f"{self.namespace}.{self.direction}"


class RemoteExecutionResponseData(RemoteExecutionMessageData):
    pass


class RemoteExecutionRequestData(RemoteExecutionMessageData):
    pass


class RemoteExecutionMessage(BaseModel):
    direction: ClassVar[Direction]
    namespace: ClassVar[Namespace]
    correlation_id: str

    @classmethod
    def message_type(cls) -> str:
        return f"{cls.namespace}.{cls.direction}"

    @property
    def result_key(self) -> str:
        return f"{self.namespace}:{self.correlation_id}"


class RemoteExecutionResponse(RemoteExecutionMessage):
    direction: ClassVar[Direction] = "response"


ResponseT = TypeVar("ResponseT", bound=RemoteExecutionResponse)


class RemoteExecutionRequest(RemoteExecutionMessage, Generic[ResponseT]):
    direction: ClassVar[Direction] = "request"

    def validate_response_type(
        self, response: RemoteExecutionMessage
    ) -> TypeGuard[ResponseT]:
        if self.namespace != response.namespace or response.direction != "response":
            raise ValueError(
                f"Expected {self.namespace}.response, but got {response.namespace}.{response.direction}"
            )
        return True


class CodeExecutionResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "code_execution"

    content: str


class CodeExecutionRequest(RemoteExecutionRequest[CodeExecutionResponse]):
    namespace: ClassVar[Namespace] = "code_execution"

    language: SupportedLanguage
    content: str


class ListFilesResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "list_files"

    filenames: list[str]


class ListFilesRequest(RemoteExecutionRequest[ListFilesResponse]):
    namespace: ClassVar[Namespace] = "list_files"

    directory: str


class GetFileContentsResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "get_file_contents"

    file_path: str
    content: str


class GetFileContentsRequest(RemoteExecutionRequest[GetFileContentsResponse]):
    namespace: ClassVar[Namespace] = "get_file_contents"

    file_path: str


class GetMatchingFilesResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "get_matching_files"

    file_paths: list[str]


class GetMatchingFilesRequest(RemoteExecutionRequest[GetMatchingFilesResponse]):
    namespace: ClassVar[Namespace] = "get_matching_files"

    search_term: str


class GetAllTrackedFilesResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "get_all_tracked_files"

    file_paths: list[str]


class GetAllTrackedFilesRequest(RemoteExecutionRequest[GetAllTrackedFilesResponse]):
    namespace: ClassVar[Namespace] = "get_all_tracked_files"


class SystemContextResponse(RemoteExecutionResponse):
    namespace: ClassVar[Namespace] = "system_context"

    exponent_txt: str | None
    system_info: SystemInfo | None


class SystemContextRequest(RemoteExecutionRequest[SystemContextResponse]):
    namespace: ClassVar[Namespace] = "system_context"


RemoteExecutionRequestType = (
    CodeExecutionRequest
    | ListFilesRequest
    | GetFileContentsRequest
    | GetMatchingFilesRequest
    | GetAllTrackedFilesRequest
    | SystemContextRequest
)

RemoteExecutionResponseType = (
    CodeExecutionResponse
    | ListFilesResponse
    | GetFileContentsResponse
    | GetMatchingFilesResponse
    | GetAllTrackedFilesResponse
    | SystemContextResponse
)


class CLIConnectedState(BaseModel):
    chat_uuid: str
    connected: bool
    last_connected_at: datetime.datetime | None
    system_info: SystemInfo | None



import json
import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager

from httpx import (
    AsyncClient,
    codes as http_status,
)

from exponent.core.remote_execution import files, git, system_context
from exponent.core.remote_execution.code_execution import execute_code
from exponent.core.remote_execution.session import (
    RemoteExecutionClientSession,
    get_session,
)
from exponent.core.remote_execution.types import (
    CLIConnectedState,
    CodeExecutionRequest,
    CreateChatResponse,
    ExecutionEndResponse,
    GetAllTrackedFilesRequest,
    GetFileContentsRequest,
    GetMatchingFilesRequest,
    HeartbeatInfo,
    ListFilesRequest,
    RemoteExecutionRequestType,
    RemoteExecutionResponse,
    StartChatRequest,
    StartChatResponse,
    SystemContextRequest,
    UseToolsConfig,
)
from exponent.core.remote_execution.utils import (
    deserialize_request_data,
    serialize_message,
)

logger = logging.getLogger(__name__)


class RemoteExecutionClient:
    def __init__(
        self,
        api_key: str,
        base_url: str,
        working_directory: str,
        session: RemoteExecutionClientSession,
    ):
        self.headers = {"API-KEY": api_key}
        self.base_url = base_url
        self.working_directory = working_directory
        self.current_session = session

        self.file_cache: files.FileCache = files.FileCache(working_directory)

    async def get_execution_requests(
        self, chat_uuid: str
    ) -> list[RemoteExecutionRequestType]:
        async with AsyncClient() as client:
            response = await client.get(
                f"{self.base_url}/api/remote_execution/{chat_uuid}/requests",
                headers=self.headers,
            )
        try:
            response_json = response.json()
        except json.JSONDecodeError as e:
            logger.error(f"Error parsing execution requests: {e}")
            logger.error(
                f"Raw response: [status-{response.status_code}] {response.text}"
            )
            raise e
        return [deserialize_request_data(result) for result in response_json]

    async def post_execution_result(
        self, chat_uuid: str, response: RemoteExecutionResponse
    ) -> None:
        async with AsyncClient() as client:
            await client.post(
                f"{self.base_url}/api/remote_execution/{chat_uuid}/result",
                headers=self.headers,
                content=serialize_message(response),
            )

    async def check_cli_execution_end_event(self, chat_uuid: str) -> bool:
        async with AsyncClient() as client:
            response = await client.get(
                f"{self.base_url}/api/remote_execution/{chat_uuid}/execution_end",
                headers=self.headers,
            )
        execution_end_response = ExecutionEndResponse(**response.json())
        return execution_end_response.execution_ended

    async def create_chat(self) -> CreateChatResponse:
        async with AsyncClient() as client:
            response = await client.post(
                f"{self.base_url}/api/remote_execution/create_chat",
                headers=self.headers,
            )
        return CreateChatResponse(**response.json())

    async def start_chat(
        self, chat_uuid: str, prompt: str, use_tools_config: UseToolsConfig
    ) -> StartChatResponse:
        async with AsyncClient() as client:
            response = await client.post(
                f"{self.base_url}/api/remote_execution/start_chat",
                headers=self.headers,
                json=StartChatRequest(
                    chat_uuid=chat_uuid,
                    prompt=prompt,
                    use_tools_config=use_tools_config,
                ).model_dump(),
                timeout=60,
            )
        return StartChatResponse(**response.json())

    async def send_heartbeat(self, chat_uuid: str) -> CLIConnectedState:
        logger.info(f"Sending heartbeat for chat_uuid {chat_uuid}")
        heartbeat_info = HeartbeatInfo(
            system_info=system_context.get_system_info(self.working_directory),
        )
        async with AsyncClient() as client:
            response = await client.post(
                f"{self.base_url}/api/remote_execution/{chat_uuid}/heartbeat",
                headers=self.headers,
                content=heartbeat_info.model_dump_json(),
                timeout=60,
            )
        if response.status_code != http_status.OK:
            raise Exception(
                f"Heartbeat failed with status code {response.status_code} and response {response.text}"
            )
        connected_state = CLIConnectedState(**response.json())
        logger.info(f"Heartbeat response: {connected_state}")
        return connected_state

    async def handle_request(
        self, request: RemoteExecutionRequestType
    ) -> RemoteExecutionResponse:
        if isinstance(request, CodeExecutionRequest):
            return await execute_code(
                request, self.current_session, working_directory=self.working_directory
            )
        elif isinstance(request, ListFilesRequest):
            return files.list_files(request)
        elif isinstance(request, GetFileContentsRequest):
            return files.get_file_contents(request, self.working_directory)
        elif isinstance(request, GetMatchingFilesRequest):
            return files.get_matching_files(request, self.file_cache)
        elif isinstance(request, SystemContextRequest):
            return system_context.get_system_context(request, self.working_directory)
        elif isinstance(request, GetAllTrackedFilesRequest):
            return git.get_all_tracked_files(request, self.working_directory)

    @staticmethod
    @asynccontextmanager
    async def session(
        api_key: str, base_url: str, working_directory: str
    ) -> AsyncGenerator["RemoteExecutionClient", None]:
        async with get_session(api_key, base_url, working_directory) as session:
            yield RemoteExecutionClient(api_key, base_url, working_directory, session)



import json
import logging
import os
from functools import lru_cache
from importlib.metadata import Distribution, PackageNotFoundError

from pydantic_settings import BaseSettings, SettingsConfigDict

logger = logging.getLogger(__name__)

# If the package is editable, we want to use:
# base_url = localhost:3000
# base_api_url = localhost:8000


def is_editable_install() -> bool:
    if os.getenv("ENVIRONMENT") == "test":
        # We should explicitly set these variables
        # in test when needed
        return False

    try:
        dist = Distribution.from_name("exponent-run")
    except PackageNotFoundError:
        logger.info("No distribution info found for exponent-run")
        return False

    direct_url = dist.read_text("direct_url.json")
    if not direct_url:
        return False

    try:
        direct_url_json = json.loads(direct_url)
    except json.JSONDecodeError:
        logger.warning("Failed to decode distribution info for exponent-run")
        return False

    pkg_is_editable = direct_url_json.get("dir_info", {}).get("editable", False)
    return bool(pkg_is_editable)


class Settings(BaseSettings):
    base_url: str = "https://exponent.run"
    base_api_url: str = "https://api.exponent.run"
    api_key: str | None = None
    log_level: str = "WARNING"

    model_config = SettingsConfigDict(
        env_prefix="EXPONENT_",
        env_file=os.path.expanduser("~/.exponent"),
        case_sensitive=False,
    )


@lru_cache(maxsize=1)
def get_settings(use_prod: bool = False, use_staging: bool = False) -> Settings:
    if is_editable_install() and not (use_prod or use_staging):
        base_url = "http://localhost:3000"
        base_api_url = "http://localhost:8000"
    elif use_staging:
        # TODO: Change this to a real staging URL?
        base_url = "https://exponent.run"
        base_api_url = "https://staging-api.exponent.run"
    else:
        base_url = "https://exponent.run"
        base_api_url = "https://api.exponent.run"

    return Settings(base_url=base_url, base_api_url=base_api_url)



import os


def execute_diff(contents: str, working_directory: str) -> str:
    try:
        # Split contents into lines and extract the filename
        lines = contents.split("\n")
        filename_line = lines[0]

        if not filename_line.startswith("filepath="):
            return "Invalid input format. The first line should start with 'filepath='."

        # Extract the relative filename and the new file content
        relative_filepath = filename_line[len("filepath=") :]
        new_content = "\n".join(lines[1:])

        # Construct the absolute path
        file_path = os.path.join(working_directory, relative_filepath)

        # Check if the directory exists, if not, create it
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        # Determine if the file exists and write the new content
        if os.path.exists(file_path):
            with open(file_path, "w") as file:
                file.write(new_content)
            return f"Modified file {relative_filepath}"
        else:
            with open(file_path, "w") as file:
                file.write(new_content)
            return f"Created file {relative_filepath}"

    except Exception as e:  # noqa: BLE001
        return f"An error occurred: {e!s}"



import os
from typing import Any

from gitignore_parser import parse_gitignore
from rapidfuzz import process

from exponent.core.remote_execution.types import (
    GetFileContentsRequest,
    GetFileContentsResponse,
    GetMatchingFilesRequest,
    GetMatchingFilesResponse,
    ListFilesRequest,
    ListFilesResponse,
)


class FileCache:
    def __init__(self, working_directory: str):
        self.working_directory = working_directory
        self._cache: list[str] | None = None

    @property
    def files(self) -> list[str]:
        if self._cache is None:
            self._cache = _file_walk(self.working_directory)
        return self._cache


def list_files(list_files_request: ListFilesRequest) -> ListFilesResponse:
    filenames = os.listdir(list_files_request.directory)
    return ListFilesResponse(
        filenames=filenames,
        correlation_id=list_files_request.correlation_id,
    )


def get_file_contents(
    get_file_contents_request: GetFileContentsRequest, working_directory: str
) -> GetFileContentsResponse:
    file_path = os.path.join(working_directory, get_file_contents_request.file_path)
    with open(file_path) as f:
        return GetFileContentsResponse(
            content=f.read(),
            file_path=get_file_contents_request.file_path,
            correlation_id=get_file_contents_request.correlation_id,
        )


def _parse_gitignore(directory: str) -> Any:
    gitignore_path = os.path.join(directory, ".gitignore")
    if os.path.isfile(gitignore_path):
        return parse_gitignore(gitignore_path)
    return None


def _file_walk(working_directory: str) -> list[str]:
    ignored_checkers = {}
    all_files = []

    for dirpath, dirnames, filenames in os.walk(working_directory):
        # Update or add new .gitignore rules when a .gitignore file is encountered
        new_ignore = _parse_gitignore(dirpath)
        if new_ignore:
            ignored_checkers[dirpath] = new_ignore

        # Check each file in the current directory
        for filename in filenames:
            file_path = os.path.join(dirpath, filename)
            # Check against all applicable .gitignore rules
            ignored = any(
                ignored_checkers[d](file_path)
                for d in ignored_checkers
                if file_path.startswith(d)
            )
            if not ignored:
                relative_path = os.path.relpath(file_path, working_directory)
                all_files.append(relative_path)

        # Update directory list in place to skip ignored directories
        dirnames[:] = [
            d
            for d in dirnames
            if not any(
                ignored_checkers[dp](os.path.join(dirpath, d))
                for dp in ignored_checkers
                if os.path.join(dirpath, d).startswith(dp)
            )
            and ".git" not in d
        ]

    return all_files


def get_matching_files(
    search_term: GetMatchingFilesRequest,
    file_cache: FileCache,
) -> GetMatchingFilesResponse:
    MAX_MATCHING_FILES = 10

    # Use rapidfuzz to find the best matching files
    matching_files = process.extract(
        search_term.search_term,
        file_cache.files,
        limit=MAX_MATCHING_FILES,
        score_cutoff=0,
    )

    file_paths: list[str] = [file for file, _, _ in matching_files]

    return GetMatchingFilesResponse(
        file_paths=file_paths, correlation_id=search_term.correlation_id
    )



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager

from exponent.core.remote_execution.languages.python import Kernel


class RemoteExecutionClientSession:
    def __init__(self, kernel: Kernel):
        self.kernel = kernel


@asynccontextmanager
async def get_session(
    api_key: str, base_url: str, working_directory: str
) -> AsyncGenerator[RemoteExecutionClientSession, None]:
    session = RemoteExecutionClientSession(Kernel())
    try:
        yield session
    except Exception as e:
        session.kernel.close()
        raise e



import getpass
import os
import platform
from pathlib import Path
from unittest import mock
from unittest.mock import mock_open

import pytest
from click.testing import CliRunner

from exponent.cli import login
from exponent.core.remote_execution import files, git
from exponent.core.remote_execution.client import RemoteExecutionClient
from exponent.core.remote_execution.languages.python import Kernel, get_python_env_info
from exponent.core.remote_execution.system_context import get_system_context
from exponent.core.remote_execution.types import (
    CodeExecutionRequest,
    CodeExecutionResponse,
    GetAllTrackedFilesRequest,
    GetAllTrackedFilesResponse,
    GetFileContentsRequest,
    GetFileContentsResponse,
    GetMatchingFilesRequest,
    GetMatchingFilesResponse,
    GitInfo,
    ListFilesRequest,
    PythonEnvInfo,
    SystemContextRequest,
    SystemContextResponse,
    SystemInfo,
)


async def test_exponent_login(cli_runner: CliRunner) -> None:
    with mock.patch("builtins.open", new_callable=mock_open):
        result = cli_runner.invoke(
            login, ["--key", "123456"], env={"EXPONENT_API_KEY": "123456"}
        )
        assert result.exit_code == 0
        assert "Saving API Key" in result.output


@mock.patch(
    "exponent.core.remote_execution.client.RemoteExecutionClient.get_execution_requests"
)
async def test_list_files(
    get_execution_requests: mock.Mock,
    default_temporary_directory: str,
) -> None:
    get_execution_requests.return_value = [
        ListFilesRequest(correlation_id="123456", directory=default_temporary_directory)
    ]

    async with RemoteExecutionClient.session(
        api_key="123456",
        base_url="https://example.com",
        working_directory=default_temporary_directory,
    ) as client:
        requests = await client.get_execution_requests(chat_uuid="123456")

        assert len(requests) == 1
        request = requests[0]
        assert isinstance(request, ListFilesRequest)
        assert request == ListFilesRequest(
            correlation_id="123456", directory=default_temporary_directory
        )

        response = files.list_files(request)
        assert response.correlation_id == "123456"
        assert sorted(response.filenames) == [
            ".git",
            "exponent.txt",
            "test1.py",
            "test2.py",
        ]


@mock.patch(
    "exponent.core.remote_execution.client.RemoteExecutionClient.get_execution_requests"
)
async def test_get_system_context(
    get_execution_requests: mock.Mock,
    default_temporary_directory: str,
    python_env_info: PythonEnvInfo,
) -> None:
    get_execution_requests.return_value = [
        SystemContextRequest(correlation_id="123456")
    ]

    async with RemoteExecutionClient.session(
        api_key="123456",
        base_url="https://example.com",
        working_directory=default_temporary_directory,
    ) as client:
        requests = await client.get_execution_requests(chat_uuid="123456")
        assert len(requests) == 1

        request = requests[0]
        assert isinstance(request, SystemContextRequest)
        assert request == SystemContextRequest(
            correlation_id="123456",
        )

        response = get_system_context(request, client.working_directory)
        assert response == SystemContextResponse(
            correlation_id="123456",
            exponent_txt="Hello, world!",
            system_info=SystemInfo(
                name=getpass.getuser(),
                cwd=default_temporary_directory,
                shell=os.environ.get("SHELL", "bash"),
                os=platform.system(),
                git=GitInfo(branch="master", remote=None),
                python_env=python_env_info,
            ),
        )


@mock.patch(
    "exponent.core.remote_execution.client.RemoteExecutionClient.get_execution_requests"
)
async def test_get_file_contents(
    get_execution_requests: mock.Mock,
    default_temporary_directory: str,
) -> None:
    get_execution_requests.return_value = [
        GetFileContentsRequest(correlation_id="123456", file_path="test1.py")
    ]

    async with RemoteExecutionClient.session(
        api_key="123456",
        base_url="https://example.com",
        working_directory=default_temporary_directory,
    ) as client:
        requests = await client.get_execution_requests(chat_uuid="123456")
        assert len(requests) == 1

        request = requests[0]
        assert isinstance(request, GetFileContentsRequest)
        assert request == GetFileContentsRequest(
            correlation_id="123456", file_path="test1.py"
        )

        response = files.get_file_contents(request, client.working_directory)
        assert response == GetFileContentsResponse(
            correlation_id="123456",
            file_path="test1.py",
            content="print('Hello, world!')",
        )


@mock.patch(
    "exponent.core.remote_execution.client.RemoteExecutionClient.get_execution_requests"
)
async def test_get_matching_files(
    get_execution_requests: mock.Mock,
    default_temporary_directory: str,
) -> None:
    get_execution_requests.return_value = [
        GetMatchingFilesRequest(correlation_id="123456", search_term="tes"),
        GetMatchingFilesRequest(correlation_id="123456", search_term="test1"),
    ]

    async with RemoteExecutionClient.session(
        api_key="123456",
        base_url="https://example.com",
        working_directory=default_temporary_directory,
    ) as client:
        EXPECTED_REQUESTS_COUNT = 2
        requests = await client.get_execution_requests(chat_uuid="123456")
        assert len(requests) == EXPECTED_REQUESTS_COUNT

        request1 = requests[0]
        assert isinstance(request1, GetMatchingFilesRequest)
        assert request1 == GetMatchingFilesRequest(
            correlation_id="123456", search_term="tes"
        )
        response1 = files.get_matching_files(request1, client.file_cache)
        assert response1 == GetMatchingFilesResponse(
            correlation_id="123456", file_paths=["test1.py", "test2.py", "exponent.txt"]
        )

        request2 = requests[1]
        assert isinstance(request2, GetMatchingFilesRequest)
        assert request2 == GetMatchingFilesRequest(
            correlation_id="123456", search_term="test1"
        )
        response2 = files.get_matching_files(request2, client.file_cache)
        assert response2 == GetMatchingFilesResponse(
            correlation_id="123456", file_paths=["test1.py", "test2.py", "exponent.txt"]
        )


@mock.patch(
    "exponent.core.remote_execution.client.RemoteExecutionClient.get_execution_requests"
)
async def test_get_matching_files_nested(
    get_execution_requests: mock.Mock,
    temporary_directory: str,
) -> None:
    # Setup test directory to look like this:
    # /folder
    #   /subfolder
    #     test1.py
    #     test2.py
    # /main
    #   core.java
    # random.tsx
    # .env
    # .gitignore (contains .env ignored)

    # Test queries for:
    # .env
    # random
    # test1
    # test2
    # .tsx
    # mjava
    import os

    # Create the necessary files and folders in the temporary directory
    os.makedirs(os.path.join(temporary_directory, "folder", "subfolder"))
    os.makedirs(os.path.join(temporary_directory, "main"))

    with open(
        os.path.join(temporary_directory, "folder", "subfolder", "test1.py"),
        "w",
    ) as f:
        f.write("# test1.py")

    with open(
        os.path.join(temporary_directory, "folder", "subfolder", "test2.py"),
        "w",
    ) as f:
        f.write("# test2.py")

    with open(os.path.join(temporary_directory, "main", "core.java"), "w") as f:
        f.write("// core.java")

    with open(os.path.join(temporary_directory, "random.tsx"), "w") as f:
        f.write("// random.tsx")

    with open(os.path.join(temporary_directory, ".env"), "w") as f:
        f.write("SECRET_KEY=123456")

    with open(os.path.join(temporary_directory, ".gitignore"), "w") as f:
        f.write(".env")

    # Test queries for various search terms
    test_cases = [
        # ("tes", ["folder/subfolder/test1.py", "folder/subfolder/test2.py"]),
        ("random", ["random.tsx"]),
        (".env", []),
        ("core", ["main/core.java"]),
        (".tsx", ["random.tsx"]),
        ("java", ["main/core.java"]),
        ("nonexistent", []),
        ("folder", ["folder/subfolder/test1.py", "folder/subfolder/test2.py"]),
        ("subfolder", ["folder/subfolder/test1.py", "folder/subfolder/test2.py"]),
        ("test1", ["folder/subfolder/test1.py"]),
        ("test2", ["folder/subfolder/test2.py"]),
    ]

    async with RemoteExecutionClient.session(
        api_key="123456",
        base_url="https://example.com",
        working_directory=temporary_directory,
    ) as client:
        # TODO: Bring back assertion that match on the matched files
        for search_term, _ in test_cases:
            get_execution_requests.return_value = [
                GetMatchingFilesRequest(
                    correlation_id="123456", search_term=search_term
                )
            ]
            requests = await client.get_execution_requests(chat_uuid="123456")
            assert len(requests) == 1

            request = requests[0]
            assert isinstance(request, GetMatchingFilesRequest)
            assert request == GetMatchingFilesRequest(
                correlation_id="123456", search_term=search_term
            )

            response = files.get_matching_files(request, client.file_cache)
            assert isinstance(response, GetMatchingFilesResponse)
            assert response.correlation_id == "123456"


@pytest.mark.skip
async def test_python_kernel() -> None:
    kernel = Kernel()
    result = await kernel.execute_code("print('Hello, world!')")
    assert result == "Hello, world!\n"


@mock.patch(
    "exponent.core.remote_execution.client.RemoteExecutionClient.get_execution_requests"
)
async def test_execute_code_diff(
    get_execution_requests: mock.Mock,
    tmpdir: Path,
) -> None:
    os.chdir(tmpdir)

    get_execution_requests.return_value = [
        CodeExecutionRequest(
            correlation_id="123456",
            language="diff",
            content='filepath=echo_server.py\nfrom fastapi import FastAPI\nfrom pydantic import BaseModel\n\napp = FastAPI()\n\nclass Message(BaseModel):\n    content: str\n\n@app.get("/echo")\ndef echo(message: Message):\n    return {"Echo": message.content}\n\nif __name__ == "__main__":\n    import uvicorn\n    uvicorn.run(app, host="127.0.0.1", port=8000)\n',
        ),
    ]

    async with RemoteExecutionClient.session(
        api_key="123456",
        base_url="https://example.com",
        working_directory=str(tmpdir),
    ) as client:
        requests = await client.get_execution_requests(chat_uuid="123456")
        assert len(requests) == 1

        request = requests[0]
        assert isinstance(request, CodeExecutionRequest)

        response = await client.handle_request(request)
        assert response == CodeExecutionResponse(
            correlation_id="123456", content="Created file echo_server.py"
        )

        with open("echo_server.py") as f:
            contents = f.read()
        assert (
            contents
            == 'from fastapi import FastAPI\nfrom pydantic import BaseModel\n\napp = FastAPI()\n\nclass Message(BaseModel):\n    content: str\n\n@app.get("/echo")\ndef echo(message: Message):\n    return {"Echo": message.content}\n\nif __name__ == "__main__":\n    import uvicorn\n    uvicorn.run(app, host="127.0.0.1", port=8000)\n'
        )


@mock.patch(
    "exponent.core.remote_execution.client.RemoteExecutionClient.get_execution_requests"
)
async def test_get_all_tracked_files(
    get_execution_requests: mock.Mock,
    default_temporary_directory: str,
) -> None:
    get_execution_requests.return_value = [
        GetAllTrackedFilesRequest(correlation_id="123456"),
    ]

    async with RemoteExecutionClient.session(
        api_key="123456",
        base_url="https://example.com",
        working_directory=default_temporary_directory,
    ) as client:
        EXPECTED_REQUESTS_COUNT = 1
        requests = await client.get_execution_requests(chat_uuid="123456")
        assert len(requests) == EXPECTED_REQUESTS_COUNT

        request1 = requests[0]
        assert isinstance(request1, GetAllTrackedFilesRequest)
        assert request1 == GetAllTrackedFilesRequest(
            correlation_id="123456",
        )
        response1 = git.get_all_tracked_files(request1, client.working_directory)
        assert response1 == GetAllTrackedFilesResponse(
            correlation_id="123456",
            file_paths=[
                "exponent.txt",
                "test1.py",
                "test2.py",
            ],
        )


def test_get_python_env_info() -> None:
    venv_info = get_python_env_info()
    assert venv_info.interpreter_path
    assert venv_info.interpreter_version
    assert venv_info.interpreter_path.endswith("python")
    assert venv_info.interpreter_version == platform.python_version()



import asyncio


async def execute_shell(code: str) -> str:
    process = await asyncio.create_subprocess_shell(
        code,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout_data, stderr_data = await process.communicate()
    decoded_stdout = stdout_data.decode()
    decoded_stderr = stderr_data.decode()
    output = []
    if decoded_stdout:
        output.append(decoded_stdout)
    if decoded_stderr:
        output.append(decoded_stderr)
    shell_output = "\n".join(output)
    return shell_output



import asyncio
import logging
import os
import sys

import click
import httpx
from dotenv import load_dotenv

from exponent.core.config import get_settings, is_editable_install
from exponent.core.remote_execution.client import RemoteExecutionClient
from exponent.core.remote_execution.types import UseToolsConfig

load_dotenv()


@click.group()
def cli() -> None:
    """Exponent CLI group."""
    set_log_level()


@cli.command()
@click.option("--key", help="Your Exponent API Key")
@click.option(
    "--prod",
    is_flag=True,
    hidden=True,
    help="Use production URLs even if in editable mode",
)
def login(key: str, prod: bool) -> None:
    settings = get_settings(use_prod=prod)
    if is_editable_install() and not prod:
        click.secho(
            "Detected local editable install, using local URLs", fg="yellow", bold=True
        )
        click.secho("- base_url=", fg="yellow", bold=True, nl=False)
        click.secho(f"{settings.base_url}", fg=(100, 200, 255), bold=False)
        click.secho("- base_api_url=", fg="yellow", bold=True, nl=False)
        click.secho(f"{settings.base_api_url}", fg=(100, 200, 255), bold=False)
        click.secho()
    if not key:
        click.echo("No API Key detected, redirecting to login...")
        click.launch(f"{settings.base_url}/")
        return
    click.echo("Saving API Key to ~/.exponent...")
    if settings.api_key and settings.api_key != key:
        click.confirm("Detected existing API Key, continue? ", default=True, abort=True)
    with open(os.path.expanduser("~/.exponent"), "a") as f:
        f.write(f"\nEXPONENT_API_KEY={key}\n")
    click.echo("API Key saved.")


@cli.command()
@click.option(
    "--chat-id",
    help="ID of an existing chat session to reconnect",
    required=False,
)
@click.option(
    "--prod",
    is_flag=True,
    hidden=True,
    help="Use production URLs even if in editable mode",
)
@click.option(
    "--staging",
    is_flag=True,
    hidden=True,
    help="Use staging URLs even if in editable mode",
)
@click.option(
    "--prompt",
    help="Start a chat with a given prompt.",
)
@click.option(
    "--benchmark",
    is_flag=True,
    help="Enable benchmarking mode",
)
def run(
    chat_id: str | None = None,
    prod: bool = False,
    staging: bool = False,
    prompt: str | None = None,
    benchmark: bool = False,
) -> None:
    chat_uuid = chat_id

    settings = get_settings(use_prod=prod, use_staging=staging)
    if is_editable_install() and not prod:
        click.secho(
            "Detected local editable install, using local URLs", fg="yellow", bold=True
        )
        click.secho("- base_url=", fg="yellow", bold=True, nl=False)
        click.secho(f"{settings.base_url}", fg=(100, 200, 255), bold=False)
        click.secho("- base_api_url=", fg="yellow", bold=True, nl=False)
        click.secho(f"{settings.base_api_url}", fg=(100, 200, 255), bold=False)
        click.secho()
    if not settings.api_key:
        click.echo("No API Key detected, redirecting to login...")
        click.launch(f"{settings.base_url}/")
        return
    else:
        api_key = settings.api_key

    loop = asyncio.get_event_loop()
    task = loop.create_task(
        start_client(
            api_key,
            settings.base_url,
            settings.base_api_url,
            chat_uuid=chat_uuid,
            prompt=prompt,
            benchmark=benchmark,
        )
    )
    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        task.cancel()
        try:
            loop.run_until_complete(task)
        except asyncio.CancelledError:
            pass


async def start_client(  # noqa: PLR0913
    api_key: str,
    base_url: str,
    base_api_url: str,
    chat_uuid: str | None = None,
    prompt: str | None = None,
    benchmark: bool = False,
) -> None:
    if benchmark is True and prompt is None:
        click.secho("Error: Benchmark mode requires a prompt.", fg="red")
        return

    current_working_directory = os.getcwd()

    if not chat_uuid:
        try:
            async with RemoteExecutionClient.session(
                api_key, base_api_url, current_working_directory
            ) as client:
                chat = await client.create_chat()
                chat_uuid = chat.chat_uuid
        except httpx.ConnectError as e:
            click.secho(f"Error: {e}", fg="red")
            return

    click.echo()
    click.secho("△ Exponent v1.0.0", fg=(180, 150, 255), bold=True)
    click.echo()
    click.echo(
        " - Link: " + click.style(f"{base_url}/chats/{chat_uuid}", fg=(100, 200, 255))
    )
    click.echo(click.style("  - Shell: /bin/zsh", fg="white"))
    click.echo()
    click.echo(click.style("✓", fg="green", bold=True) + " Ready in 1401ms")

    # Open the chat in the browser
    if not benchmark and not prompt and not chat_uuid:
        click.launch(f"{base_url}/chats/{chat_uuid}")

    use_tools_config = UseToolsConfig()
    async with RemoteExecutionClient.session(
        api_key, base_api_url, current_working_directory
    ) as client:
        if benchmark:
            assert prompt is not None
            await asyncio.gather(
                start_chat(
                    client, chat_uuid, prompt, use_tools_config=use_tools_config
                ),
                run_execution_client(client, chat_uuid),
                benchmark_mode_exit(client, chat_uuid),
                heartbeat_thread(client, chat_uuid),
            )
        elif prompt:
            await asyncio.gather(
                start_chat(
                    client, chat_uuid, prompt, use_tools_config=use_tools_config
                ),
                run_execution_client(client, chat_uuid),
                heartbeat_thread(client, chat_uuid),
            )
        else:
            await asyncio.gather(
                run_execution_client(client, chat_uuid),
                heartbeat_thread(client, chat_uuid),
            )


async def run_execution_client(client: RemoteExecutionClient, chat_uuid: str) -> None:
    while True:
        execution_requests = await client.get_execution_requests(chat_uuid)
        for execution_request in execution_requests:
            click.echo(f"Handling {execution_request.namespace} request:")
            click.echo(f"  - {execution_request}")
            execution_response = await client.handle_request(execution_request)
            click.echo(f"Posting {execution_request.namespace} response.")
            await client.post_execution_result(chat_uuid, execution_response)
        await asyncio.sleep(0.2)


async def start_chat(
    client: RemoteExecutionClient,
    chat_uuid: str,
    prompt: str,
    use_tools_config: UseToolsConfig,
) -> None:
    click.secho("Starting chat...")
    await client.start_chat(chat_uuid, prompt, use_tools_config=use_tools_config)
    click.secho("Chat started. Open the link to join the chat.")


async def benchmark_mode_exit(client: RemoteExecutionClient, chat_uuid: str) -> None:
    while True:
        await asyncio.sleep(5)
        if await client.check_cli_execution_end_event(chat_uuid):
            sys.exit(0)


async def heartbeat_thread(client: RemoteExecutionClient, chat_uuid: str) -> None:
    while True:
        await client.send_heartbeat(chat_uuid)
        await asyncio.sleep(1)


def set_log_level() -> None:
    settings = get_settings()
    logging.basicConfig(level=getattr(logging, settings.log_level), stream=sys.stdout)


if __name__ == "__main__":
    cli()



import asyncio
import os
import tempfile
from collections.abc import Generator
from unittest.mock import patch

import pygit2
import pytest
from click.testing import CliRunner

from exponent.core.remote_execution.languages import python
from exponent.core.remote_execution.types import PythonEnvInfo
from exponent.tests.utils import create_commit, stage_file


@pytest.fixture(scope="session")
def event_loop() -> Generator[asyncio.AbstractEventLoop, None, None]:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.close()


@pytest.fixture
def cli_runner() -> CliRunner:
    return CliRunner()


@pytest.fixture(scope="session", autouse=True)
async def fixture_setup_env() -> None:
    session_mpatch = pytest.MonkeyPatch()
    session_mpatch.setenv("EXPONENT_API_KEY", "123456")
    session_mpatch.setenv("EXPONENT_BASE_URL", "https://exponent.run")
    session_mpatch.setenv("EXPONENT_API_BASE_URL", "https://api.exponent.run")


@pytest.fixture(scope="function")
def temporary_directory() -> Generator[str, None, None]:
    with tempfile.TemporaryDirectory() as test_directory:
        yield test_directory


@pytest.fixture(scope="function")
def default_temporary_directory(temporary_directory: str) -> Generator[str, None, None]:
    with open(os.path.join(temporary_directory, "test1.py"), "w") as f:
        f.write("print('Hello, world!')")

    with open(os.path.join(temporary_directory, "test2.py"), "w") as f:
        f.write("print('Hello, world!')")

    with open(os.path.join(temporary_directory, "exponent.txt"), "w") as f:
        f.write("Hello, world!")

    repo = pygit2.init_repository(temporary_directory, bare=False)

    stage_file(repo, "test1.py")
    stage_file(repo, "test2.py")
    stage_file(repo, "exponent.txt")
    create_commit(repo, "Initial commit", "Tester McTesterface", "test@example.com")

    yield temporary_directory


@pytest.fixture(scope="function", autouse=True)
def mock_cli_heartbeat() -> Generator[None, None, None]:
    patcher = patch(
        "exponent.core.remote_execution.client.RemoteExecutionClient.send_heartbeat"
    )
    patcher.start()
    yield
    patcher.stop()


@pytest.fixture(name="python_env_info", scope="function", autouse=True)
def fixture_python_env_info(
    monkeypatch: pytest.MonkeyPatch,
) -> PythonEnvInfo:
    env_info = PythonEnvInfo(
        interpreter_path="python",
        interpreter_version="3.11.0",
    )

    def get_python_env_info() -> PythonEnvInfo:
        return env_info

    monkeypatch.setattr(
        python,
        "get_python_env_info",
        get_python_env_info,
    )

    return env_info



import pygit2

from exponent.core.remote_execution.types import (
    GetAllTrackedFilesRequest,
    GetAllTrackedFilesResponse,
    GitInfo,
)


def get_all_tracked_files(
    request: GetAllTrackedFilesRequest,
    working_directory: str,
) -> GetAllTrackedFilesResponse:
    return GetAllTrackedFilesResponse(
        correlation_id=request.correlation_id,
        file_paths=get_all_tracked_git_file_paths(working_directory),
    )


def get_all_tracked_git_file_paths(working_directory: str) -> list[str]:
    repo = pygit2.Repository(working_directory)
    return sorted([entry.path for entry in repo.index])


def get_git_info(working_directory: str) -> GitInfo | None:
    try:
        repo = pygit2.Repository(working_directory)
    except pygit2.GitError:
        return None

    return GitInfo(
        branch=_get_git_branch(repo),
        remote=_get_git_remote(repo),
    )


def _get_git_remote(repo: pygit2.Repository) -> str | None:
    if repo.remotes:
        return str(repo.remotes[0].url)
    return None


def _get_git_branch(repo: pygit2.Repository) -> str:
    return str(repo.head.shorthand)
